From 194f07aa7f6abdd4ade3759907a9c84b9bc82814 Mon Sep 17 00:00:00 2001
From: Alexander Korotkov <akorotkov@postgresql.org>
Date: Sun, 20 Nov 2022 03:29:09 +0300
Subject: [PATCH] USER SET parameters for pg_db_role_setting

Specifies that variable should be set on behalf of ordinal role.
That lets ordinal role set placeholder variables, which permission
requirements is not known yet.
The value set wouldn't be used if variable finally appear to require
superuser privileges.
---
 doc/src/sgml/ref/alter_database.sgml          |  15 ++-
 doc/src/sgml/ref/alter_role.sgml              |  22 ++-
 src/backend/catalog/pg_db_role_setting.c      |   4 +-
 src/backend/commands/functioncmds.c           |   2 +-
 src/backend/parser/gram.y                     |  20 +++
 src/backend/utils/misc/guc.c                  | 125 +++++++++++++-----
 src/backend/utils/misc/guc_funcs.c            |  12 +-
 src/bin/pg_dump/dumputils.c                   |  18 ++-
 src/include/common/guc-common.h               |  31 +++++
 src/include/nodes/parsenodes.h                |   1 +
 src/include/utils/guc.h                       |   4 +-
 src/test/modules/Makefile                     |   1 +
 src/test/modules/meson.build                  |   1 +
 .../test_pg_db_role_setting/.gitignore        |   4 +
 .../modules/test_pg_db_role_setting/Makefile  |  29 ++++
 .../expected/test_pg_db_role_setting.out      | 120 +++++++++++++++++
 .../test_pg_db_role_setting/meson.build       |  35 +++++
 .../sql/test_pg_db_role_setting.sql           |  55 ++++++++
 .../test_pg_db_role_setting--1.0.sql          |   7 +
 .../test_pg_db_role_setting.c                 |  57 ++++++++
 .../test_pg_db_role_setting.control           |   7 +
 21 files changed, 526 insertions(+), 44 deletions(-)
 create mode 100644 src/include/common/guc-common.h
 create mode 100644 src/test/modules/test_pg_db_role_setting/.gitignore
 create mode 100644 src/test/modules/test_pg_db_role_setting/Makefile
 create mode 100644 src/test/modules/test_pg_db_role_setting/expected/test_pg_db_role_setting.out
 create mode 100644 src/test/modules/test_pg_db_role_setting/meson.build
 create mode 100644 src/test/modules/test_pg_db_role_setting/sql/test_pg_db_role_setting.sql
 create mode 100644 src/test/modules/test_pg_db_role_setting/test_pg_db_role_setting--1.0.sql
 create mode 100644 src/test/modules/test_pg_db_role_setting/test_pg_db_role_setting.c
 create mode 100644 src/test/modules/test_pg_db_role_setting/test_pg_db_role_setting.control

diff --git a/doc/src/sgml/ref/alter_database.sgml b/doc/src/sgml/ref/alter_database.sgml
index 89ed261b4c2..697db7b720f 100644
--- a/doc/src/sgml/ref/alter_database.sgml
+++ b/doc/src/sgml/ref/alter_database.sgml
@@ -37,7 +37,7 @@ ALTER DATABASE <replaceable class="parameter">name</replaceable> SET TABLESPACE
 
 ALTER DATABASE <replaceable class="parameter">name</replaceable> REFRESH COLLATION VERSION
 
-ALTER DATABASE <replaceable class="parameter">name</replaceable> SET <replaceable>configuration_parameter</replaceable> { TO | = } { <replaceable>value</replaceable> | DEFAULT }
+ALTER DATABASE <replaceable class="parameter">name</replaceable> SET <replaceable>configuration_parameter</replaceable> { TO | = } { <replaceable>value</replaceable> | <replaceable>value</replaceable> USER SET | DEFAULT }
 ALTER DATABASE <replaceable class="parameter">name</replaceable> SET <replaceable>configuration_parameter</replaceable> FROM CURRENT
 ALTER DATABASE <replaceable class="parameter">name</replaceable> RESET <replaceable>configuration_parameter</replaceable>
 ALTER DATABASE <replaceable class="parameter">name</replaceable> RESET ALL
@@ -206,6 +206,19 @@ ALTER DATABASE <replaceable class="parameter">name</replaceable> RESET ALL
        </para>
       </listitem>
      </varlistentry>
+
+     <varlistentry>
+      <term><literal>USER SET</literal></term>
+      <listitem>
+       <para>
+        Specifies that variable should be set on behalf of ordinal role.
+        That lets ordinal role set placeholder variables, which permission
+        requirements is not known yet; see <xref linkend="runtime-config-custom"/>.
+        The value set wouldn't be used if variable finally appear to require
+        superuser privileges.
+       </para>
+      </listitem>
+     </varlistentry>
   </variablelist>
  </refsect1>
 
diff --git a/doc/src/sgml/ref/alter_role.sgml b/doc/src/sgml/ref/alter_role.sgml
index 5aa5648ae7b..0c51d1d0561 100644
--- a/doc/src/sgml/ref/alter_role.sgml
+++ b/doc/src/sgml/ref/alter_role.sgml
@@ -38,7 +38,7 @@ ALTER ROLE <replaceable class="parameter">role_specification</replaceable> [ WIT
 
 ALTER ROLE <replaceable class="parameter">name</replaceable> RENAME TO <replaceable>new_name</replaceable>
 
-ALTER ROLE { <replaceable class="parameter">role_specification</replaceable> | ALL } [ IN DATABASE <replaceable class="parameter">database_name</replaceable> ] SET <replaceable>configuration_parameter</replaceable> { TO | = } { <replaceable>value</replaceable> | DEFAULT }
+ALTER ROLE { <replaceable class="parameter">role_specification</replaceable> | ALL } [ IN DATABASE <replaceable class="parameter">database_name</replaceable> ] SET <replaceable>configuration_parameter</replaceable> { TO | = } { <replaceable>value</replaceable> | <replaceable>value</replaceable> USER SET | DEFAULT }
 ALTER ROLE { <replaceable class="parameter">role_specification</replaceable> | ALL } [ IN DATABASE <replaceable class="parameter">database_name</replaceable> ] SET <replaceable>configuration_parameter</replaceable> FROM CURRENT
 ALTER ROLE { <replaceable class="parameter">role_specification</replaceable> | ALL } [ IN DATABASE <replaceable class="parameter">database_name</replaceable> ] RESET <replaceable>configuration_parameter</replaceable>
 ALTER ROLE { <replaceable class="parameter">role_specification</replaceable> | ALL } [ IN DATABASE <replaceable class="parameter">database_name</replaceable> ] RESET ALL
@@ -234,6 +234,19 @@ ALTER ROLE { <replaceable class="parameter">role_specification</replaceable> | A
        </para>
       </listitem>
      </varlistentry>
+
+     <varlistentry>
+      <term><literal>USER SET</literal></term>
+      <listitem>
+       <para>
+        Specifies that variable should be set on behalf of ordinal role.
+        That lets ordinal role set placeholder variables, which permission
+        requirements is not known yet; see <xref linkend="runtime-config-custom"/>.
+        The value set wouldn't be used if variable finally appear to require
+        superuser privileges.
+       </para>
+      </listitem>
+     </varlistentry>
     </variablelist>
  </refsect1>
 
@@ -329,6 +342,13 @@ ALTER ROLE worker_bee SET maintenance_work_mem = 100000;
 
 <programlisting>
 ALTER ROLE fred IN DATABASE devel SET client_min_messages = DEBUG;
+</programlisting></para>
+
+  <para>
+   Give a role a non-default placeholder setting on behalf of ordinal user.
+
+<programlisting>
+ALTER ROLE fred SET my.param = 'value' USER SET;
 </programlisting></para>
  </refsect1>
 
diff --git a/src/backend/catalog/pg_db_role_setting.c b/src/backend/catalog/pg_db_role_setting.c
index 42387f4e304..4b9a39a953d 100644
--- a/src/backend/catalog/pg_db_role_setting.c
+++ b/src/backend/catalog/pg_db_role_setting.c
@@ -115,7 +115,7 @@ AlterSetting(Oid databaseid, Oid roleid, VariableSetStmt *setstmt)
 
 		/* Update (valuestr is NULL in RESET cases) */
 		if (valuestr)
-			a = GUCArrayAdd(a, setstmt->name, valuestr);
+			a = GUCArrayAdd(a, setstmt->name, valuestr, setstmt->user_set);
 		else
 			a = GUCArrayDelete(a, setstmt->name);
 
@@ -141,7 +141,7 @@ AlterSetting(Oid databaseid, Oid roleid, VariableSetStmt *setstmt)
 
 		memset(nulls, false, sizeof(nulls));
 
-		a = GUCArrayAdd(NULL, setstmt->name, valuestr);
+		a = GUCArrayAdd(NULL, setstmt->name, valuestr, setstmt->user_set);
 
 		values[Anum_pg_db_role_setting_setdatabase - 1] =
 			ObjectIdGetDatum(databaseid);
diff --git a/src/backend/commands/functioncmds.c b/src/backend/commands/functioncmds.c
index 1f820c93e96..950fd28badc 100644
--- a/src/backend/commands/functioncmds.c
+++ b/src/backend/commands/functioncmds.c
@@ -662,7 +662,7 @@ update_proconfig_value(ArrayType *a, List *set_items)
 			char	   *valuestr = ExtractSetVariableArgs(sstmt);
 
 			if (valuestr)
-				a = GUCArrayAdd(a, sstmt->name, valuestr);
+				a = GUCArrayAdd(a, sstmt->name, valuestr, sstmt->user_set);
 			else				/* RESET */
 				a = GUCArrayDelete(a, sstmt->name);
 		}
diff --git a/src/backend/parser/gram.y b/src/backend/parser/gram.y
index 737bd2d06d5..c38398d8528 100644
--- a/src/backend/parser/gram.y
+++ b/src/backend/parser/gram.y
@@ -1622,6 +1622,26 @@ generic_set:
 					n->args = $3;
 					$$ = n;
 				}
+			| var_name TO var_list USER SET
+				{
+					VariableSetStmt *n = makeNode(VariableSetStmt);
+
+					n->kind = VAR_SET_VALUE;
+					n->name = $1;
+					n->args = $3;
+					n->user_set = true;
+					$$ = n;
+				}
+			| var_name '=' var_list USER SET
+				{
+					VariableSetStmt *n = makeNode(VariableSetStmt);
+
+					n->kind = VAR_SET_VALUE;
+					n->name = $1;
+					n->args = $3;
+					n->user_set = true;
+					$$ = n;
+				}
 			| var_name TO DEFAULT
 				{
 					VariableSetStmt *n = makeNode(VariableSetStmt);
diff --git a/src/backend/utils/misc/guc.c b/src/backend/utils/misc/guc.c
index 117a2d26a0e..d1564c7284e 100644
--- a/src/backend/utils/misc/guc.c
+++ b/src/backend/utils/misc/guc.c
@@ -224,7 +224,6 @@ static bool reporting_enabled;	/* true to enable GUC_REPORT */
 
 static int	GUCNestLevel = 0;	/* 1 when in main transaction */
 
-
 static int	guc_var_compare(const void *a, const void *b);
 static uint32 guc_name_hash(const void *key, Size keysize);
 static int	guc_name_match(const void *key1, const void *key2, Size keysize);
@@ -244,7 +243,7 @@ static void reapply_stacked_values(struct config_generic *variable,
 								   GucContext curscontext, GucSource cursource,
 								   Oid cursrole);
 static bool validate_option_array_item(const char *name, const char *value,
-									   bool skipIfNoPermissions);
+									   bool user_set, bool skipIfNoPermissions);
 static void write_auto_conf_file(int fd, const char *filename, ConfigVariable *head);
 static void replace_auto_config_value(ConfigVariable **head_p, ConfigVariable **tail_p,
 									  const char *name, const char *value);
@@ -6159,13 +6158,16 @@ RestoreGUCState(void *gucstate)
 
 /*
  * A little "long argument" simulation, although not quite GNU
- * compliant. Takes a string of the form "some-option=some value" and
- * returns name = "some_option" and value = "some value" in palloc'ed
- * storage. Note that '-' is converted to '_' in the option name. If
- * there is no '=' in the input string then value will be NULL.
+ * compliant. Takes a string of the form "some-option=some value" or
+ * "some-option(u)=some value" and returns name = "some_option" and
+ * value = "some value" in palloc'ed storage. If user_set is not null then
+ * the presence of "(u)" flag is stored there. Note that '-' is converted
+ * to '_' in the option name. If there is no '=' in the input string then
+ * value will be NULL.
  */
-void
-ParseLongOption(const char *string, char **name, char **value)
+static void
+ParseLongOptionInternal(const char *string, char **name, char **value,
+						bool *user_set)
 {
 	size_t		equal_pos;
 	char	   *cp;
@@ -6176,25 +6178,41 @@ ParseLongOption(const char *string, char **name, char **value)
 
 	equal_pos = strcspn(string, "=");
 
-	if (string[equal_pos] == '=')
+	if (GUC_ARRAY_IS_USERSET_SIGN_BEFORE(string, string + equal_pos))
+	{
+		*name = palloc(equal_pos - GUC_ARRAY_USERSET_SIGN_LEN + 1);
+		strlcpy(*name, string, equal_pos - GUC_ARRAY_USERSET_SIGN_LEN + 1);
+		if (user_set)
+			*user_set = true;
+	}
+	else
 	{
 		*name = palloc(equal_pos + 1);
 		strlcpy(*name, string, equal_pos + 1);
+		if (user_set)
+			*user_set = false;
+	}
 
+	if (string[equal_pos] == '=')
 		*value = pstrdup(&string[equal_pos + 1]);
-	}
 	else
-	{
-		/* no equal sign in string */
-		*name = pstrdup(string);
 		*value = NULL;
-	}
 
 	for (cp = *name; *cp; cp++)
 		if (*cp == '-')
 			*cp = '_';
 }
 
+/*
+ * The exported version of ParseLongOptionInternal().  Doesn't need user_set
+ * argument since no external users need it.
+ */
+void
+ParseLongOption(const char *string, char **name, char **value)
+{
+	ParseLongOptionInternal(string, name, value, NULL);
+}
+
 
 /*
  * Handle options fetched from pg_db_role_setting.setconfig,
@@ -6220,6 +6238,7 @@ ProcessGUCArray(ArrayType *array,
 		char	   *s;
 		char	   *name;
 		char	   *value;
+		bool		user_set;
 
 		d = array_ref(array, 1, &i,
 					  -1 /* varlenarray */ ,
@@ -6233,7 +6252,7 @@ ProcessGUCArray(ArrayType *array,
 
 		s = TextDatumGetCString(d);
 
-		ParseLongOption(s, &name, &value);
+		ParseLongOptionInternal(s, &name, &value, &user_set);
 		if (!value)
 		{
 			ereport(WARNING,
@@ -6244,9 +6263,19 @@ ProcessGUCArray(ArrayType *array,
 			continue;
 		}
 
-		(void) set_config_option(name, value,
-								 context, source,
-								 action, true, 0, false);
+		/*
+		 * USER SET values are appliciable only for PGC_USERSET parameters.
+		 * We use InvalidOid as role in order to evade possible privileges of
+		 * the current user.
+		 */
+		if (!user_set)
+			(void) set_config_option(name, value,
+									 context, source,
+									 action, true, 0, false);
+		else
+			(void) set_config_option_ext(name, value,
+										 PGC_USERSET, source, InvalidOid,
+										 action, true, 0, false);
 
 		pfree(name);
 		pfree(value);
@@ -6260,7 +6289,8 @@ ProcessGUCArray(ArrayType *array,
  * to indicate the current table entry is NULL.
  */
 ArrayType *
-GUCArrayAdd(ArrayType *array, const char *name, const char *value)
+GUCArrayAdd(ArrayType *array, const char *name, const char *value,
+			bool user_set)
 {
 	struct config_generic *record;
 	Datum		datum;
@@ -6271,7 +6301,7 @@ GUCArrayAdd(ArrayType *array, const char *name, const char *value)
 	Assert(value);
 
 	/* test if the option is valid and we're allowed to set it */
-	(void) validate_option_array_item(name, value, false);
+	(void) validate_option_array_item(name, value, user_set, false);
 
 	/* normalize name (converts obsolete GUC names to modern spellings) */
 	record = find_option(name, false, true, WARNING);
@@ -6279,7 +6309,11 @@ GUCArrayAdd(ArrayType *array, const char *name, const char *value)
 		name = record->name;
 
 	/* build new item for array */
-	newval = psprintf("%s=%s", name, value);
+	if (user_set)
+		newval = psprintf("%s" GUC_ARRAY_USERSET_SIGN "=%s",
+						  name, value);
+	else
+		newval = psprintf("%s=%s", name, value);
 	datum = CStringGetTextDatum(newval);
 
 	if (array)
@@ -6309,9 +6343,17 @@ GUCArrayAdd(ArrayType *array, const char *name, const char *value)
 				continue;
 			current = TextDatumGetCString(d);
 
-			/* check for match up through and including '=' */
-			if (strncmp(current, newval, strlen(name) + 1) == 0)
+			/* check for the name match */
+			if (strncmp(current, newval, strlen(name)) == 0 &&
+				GUC_ARRAY_IS_NAME_BORDER(current + strlen(name)))
 			{
+				/*
+				 * Recheck permissons if we found an option without USER SET
+				 * flag while we're setting an optionn with USER SET flag.
+				 */
+				if (current[strlen(name)] == '=' && user_set)
+					(void) validate_option_array_item(name, value,
+													  false, false);
 				index = i;
 				break;
 			}
@@ -6347,9 +6389,6 @@ GUCArrayDelete(ArrayType *array, const char *name)
 
 	Assert(name);
 
-	/* test if the option is valid and we're allowed to set it */
-	(void) validate_option_array_item(name, NULL, false);
-
 	/* normalize name (converts obsolete GUC names to modern spellings) */
 	record = find_option(name, false, true, WARNING);
 	if (record)
@@ -6379,9 +6418,15 @@ GUCArrayDelete(ArrayType *array, const char *name)
 		val = TextDatumGetCString(d);
 
 		/* ignore entry if it's what we want to delete */
-		if (strncmp(val, name, strlen(name)) == 0
-			&& val[strlen(name)] == '=')
+		if (strncmp(val, name, strlen(name)) == 0 &&
+			GUC_ARRAY_IS_NAME_BORDER(val + strlen(name)))
+		{
+			/* test if the option is valid and we're allowed to set it */
+			(void) validate_option_array_item(name, NULL,
+											  GUC_ARRAY_IS_USERSET_SIGN(val + strlen(name)),
+											  false);
 			continue;
+		}
 
 		/* else add it to the output array */
 		if (newarray)
@@ -6431,6 +6476,7 @@ GUCArrayReset(ArrayType *array)
 		char	   *val;
 		char	   *eqsgn;
 		bool		isnull;
+		bool		user_set = false;
 
 		d = array_ref(array, 1, &i,
 					  -1 /* varlenarray */ ,
@@ -6443,10 +6489,18 @@ GUCArrayReset(ArrayType *array)
 		val = TextDatumGetCString(d);
 
 		eqsgn = strchr(val, '=');
-		*eqsgn = '\0';
+		if (GUC_ARRAY_IS_USERSET_SIGN_BEFORE(val, eqsgn))
+		{
+			*(eqsgn - GUC_ARRAY_USERSET_SIGN_LEN) = '\0';
+			user_set = true;
+		}
+		else
+		{
+			*eqsgn = '\0';
+		}
 
 		/* skip if we have permission to delete it */
-		if (validate_option_array_item(val, NULL, true))
+		if (validate_option_array_item(val, NULL, user_set, true))
 			continue;
 
 		/* else add it to the output array */
@@ -6472,15 +6526,16 @@ GUCArrayReset(ArrayType *array)
  * Validate a proposed option setting for GUCArrayAdd/Delete/Reset.
  *
  * name is the option name.  value is the proposed value for the Add case,
- * or NULL for the Delete/Reset cases.  If skipIfNoPermissions is true, it's
- * not an error to have no permissions to set the option.
+ * or NULL for the Delete/Reset cases.  user_set indicates this is the USER SET
+ * option.  If skipIfNoPermissions is true, it's not an error to have no
+ * permissions to set the option.
  *
  * Returns true if OK, false if skipIfNoPermissions is true and user does not
  * have permission to change this option (all other error cases result in an
  * error being thrown).
  */
 static bool
-validate_option_array_item(const char *name, const char *value,
+validate_option_array_item(const char *name, const char *value, bool user_set,
 						   bool skipIfNoPermissions)
 
 {
@@ -6516,8 +6571,10 @@ validate_option_array_item(const char *name, const char *value,
 	{
 		/*
 		 * We cannot do any meaningful check on the value, so only permissions
-		 * are useful to check.
+		 * are useful to check.  USER SET options are always allowed.
 		 */
+		if (user_set)
+			return true;
 		if (superuser() ||
 			pg_parameter_aclcheck(name, GetUserId(), ACL_SET) == ACLCHECK_OK)
 			return true;
diff --git a/src/backend/utils/misc/guc_funcs.c b/src/backend/utils/misc/guc_funcs.c
index 108b3bd1290..963921710cd 100644
--- a/src/backend/utils/misc/guc_funcs.c
+++ b/src/backend/utils/misc/guc_funcs.c
@@ -166,12 +166,22 @@ ExecSetVariableStmt(VariableSetStmt *stmt, bool isTopLevel)
 char *
 ExtractSetVariableArgs(VariableSetStmt *stmt)
 {
+
 	switch (stmt->kind)
 	{
 		case VAR_SET_VALUE:
 			return flatten_set_variable_args(stmt->name, stmt->args);
 		case VAR_SET_CURRENT:
-			return GetConfigOptionByName(stmt->name, NULL, false);
+		{
+			struct config_generic *record;
+			char	   *result;
+
+			result = GetConfigOptionByName(stmt->name, NULL, false);
+			record = find_option(stmt->name, false, false, ERROR);
+			stmt->user_set = (record->scontext == PGC_USERSET);
+
+			return result;
+		}
 		default:
 			return NULL;
 	}
diff --git a/src/bin/pg_dump/dumputils.c b/src/bin/pg_dump/dumputils.c
index 6e501a54138..4316232d1d1 100644
--- a/src/bin/pg_dump/dumputils.c
+++ b/src/bin/pg_dump/dumputils.c
@@ -16,6 +16,7 @@
 
 #include <ctype.h>
 
+#include "common/guc-common.h"
 #include "dumputils.h"
 #include "fe_utils/string_utils.h"
 
@@ -804,8 +805,8 @@ SplitGUCList(char *rawstring, char separator,
 /*
  * Helper function for dumping "ALTER DATABASE/ROLE SET ..." commands.
  *
- * Parse the contents of configitem (a "name=value" string), wrap it in
- * a complete ALTER command, and append it to buf.
+ * Parse the contents of configitem (a "name=value" or "name(u)=value" string),
+ * wrap it in a complete ALTER command, and append it to buf.
  *
  * type is DATABASE or ROLE, and name is the name of the database or role.
  * If we need an "IN" clause, type2 and name2 similarly define what to put
@@ -820,6 +821,7 @@ makeAlterConfigCommand(PGconn *conn, const char *configitem,
 {
 	char	   *mine;
 	char	   *pos;
+	bool		user_set = false;
 
 	/* Parse the configitem.  If we can't find an "=", silently do nothing. */
 	mine = pg_strdup(configitem);
@@ -829,7 +831,13 @@ makeAlterConfigCommand(PGconn *conn, const char *configitem,
 		pg_free(mine);
 		return;
 	}
-	*pos++ = '\0';
+	if (GUC_ARRAY_IS_USERSET_SIGN_BEFORE(mine, pos))
+	{
+		user_set = true;
+		*(pos - GUC_ARRAY_USERSET_SIGN_LEN) = '\0';
+	}
+	else
+		*pos++ = '\0';
 
 	/* Build the command, with suitable quoting for everything. */
 	appendPQExpBuffer(buf, "ALTER %s %s ", type, fmtId(name));
@@ -872,6 +880,10 @@ makeAlterConfigCommand(PGconn *conn, const char *configitem,
 	else
 		appendStringLiteralConn(buf, pos, conn);
 
+	/* Add USER SET flag if specified in the string */
+	if (user_set)
+		appendPQExpBufferStr(buf, "USER SET;\n");
+
 	appendPQExpBufferStr(buf, ";\n");
 
 	pg_free(mine);
diff --git a/src/include/common/guc-common.h b/src/include/common/guc-common.h
new file mode 100644
index 00000000000..8b31953789e
--- /dev/null
+++ b/src/include/common/guc-common.h
@@ -0,0 +1,31 @@
+/*-------------------------------------------------------------------------
+ *
+ * guc-common.h
+ *	  Common declarations for Grand Unified Configuration.
+ *
+ *
+ * Portions Copyright (c) 1996-2022, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1994, Regents of the University of California
+ *
+ * src/include/common/guc-common.h
+ *
+ *-------------------------------------------------------------------------
+ */
+#ifndef GUC_COMMON_H
+#define GUC_COMMON_H
+
+/*
+ * The designator of USER SET value in GUC array.
+ */
+#define GUC_ARRAY_USERSET_SIGN "(u)"
+#define GUC_ARRAY_USERSET_SIGN_LEN \
+	(sizeof(GUC_ARRAY_USERSET_SIGN) - 1)
+#define GUC_ARRAY_IS_USERSET_SIGN(s) \
+	(strncmp((s), GUC_ARRAY_USERSET_SIGN, GUC_ARRAY_USERSET_SIGN_LEN) == 0)
+#define GUC_ARRAY_IS_USERSET_SIGN_BEFORE(start, eqsign) \
+	((eqsign) - (start) >= GUC_ARRAY_USERSET_SIGN_LEN && \
+	 GUC_ARRAY_IS_USERSET_SIGN((eqsign) - GUC_ARRAY_USERSET_SIGN_LEN))
+#define GUC_ARRAY_IS_NAME_BORDER(s) \
+	((*(s)) == '=' || GUC_ARRAY_IS_USERSET_SIGN(s))
+
+#endif							/* GUC_COMMON_H */
diff --git a/src/include/nodes/parsenodes.h b/src/include/nodes/parsenodes.h
index 7e7ad3f7e47..d48acda7c7b 100644
--- a/src/include/nodes/parsenodes.h
+++ b/src/include/nodes/parsenodes.h
@@ -2228,6 +2228,7 @@ typedef struct VariableSetStmt
 	char	   *name;			/* variable to be set */
 	List	   *args;			/* List of A_Const nodes */
 	bool		is_local;		/* SET LOCAL? */
+	bool		user_set;
 } VariableSetStmt;
 
 /* ----------------------
diff --git a/src/include/utils/guc.h b/src/include/utils/guc.h
index b3aaff9665b..9802973f086 100644
--- a/src/include/utils/guc.h
+++ b/src/include/utils/guc.h
@@ -12,6 +12,7 @@
 #ifndef GUC_H
 #define GUC_H
 
+#include "common/guc-common.h"
 #include "nodes/parsenodes.h"
 #include "tcop/dest.h"
 #include "utils/array.h"
@@ -393,7 +394,8 @@ extern char *GetConfigOptionByName(const char *name, const char **varname,
 
 extern void ProcessGUCArray(ArrayType *array,
 							GucContext context, GucSource source, GucAction action);
-extern ArrayType *GUCArrayAdd(ArrayType *array, const char *name, const char *value);
+extern ArrayType *GUCArrayAdd(ArrayType *array, const char *name,
+							  const char *value, bool user_set);
 extern ArrayType *GUCArrayDelete(ArrayType *array, const char *name);
 extern ArrayType *GUCArrayReset(ArrayType *array);
 
diff --git a/src/test/modules/Makefile b/src/test/modules/Makefile
index 7b3f2929652..ef724703d46 100644
--- a/src/test/modules/Makefile
+++ b/src/test/modules/Makefile
@@ -24,6 +24,7 @@ SUBDIRS = \
 		  test_misc \
 		  test_oat_hooks \
 		  test_parser \
+		  test_pg_db_role_settings \
 		  test_pg_dump \
 		  test_predtest \
 		  test_rbtree \
diff --git a/src/test/modules/meson.build b/src/test/modules/meson.build
index c2e5f5ffd5a..c152723a277 100644
--- a/src/test/modules/meson.build
+++ b/src/test/modules/meson.build
@@ -18,6 +18,7 @@ subdir('test_lfind')
 subdir('test_misc')
 subdir('test_oat_hooks')
 subdir('test_parser')
+subdir('test_pg_db_role_settings')
 subdir('test_pg_dump')
 subdir('test_predtest')
 subdir('test_rbtree')
diff --git a/src/test/modules/test_pg_db_role_setting/.gitignore b/src/test/modules/test_pg_db_role_setting/.gitignore
new file mode 100644
index 00000000000..5dcb3ff9723
--- /dev/null
+++ b/src/test/modules/test_pg_db_role_setting/.gitignore
@@ -0,0 +1,4 @@
+# Generated subdirectories
+/log/
+/results/
+/tmp_check/
diff --git a/src/test/modules/test_pg_db_role_setting/Makefile b/src/test/modules/test_pg_db_role_setting/Makefile
new file mode 100644
index 00000000000..aacd78f74c5
--- /dev/null
+++ b/src/test/modules/test_pg_db_role_setting/Makefile
@@ -0,0 +1,29 @@
+# src/test/modules/test_pg_db_role_setting/Makefile
+
+MODULE_big = test_pg_db_role_setting
+OBJS = \
+	$(WIN32RES) \
+	test_pg_db_role_setting.o
+EXTENSION = test_pg_db_role_setting
+DATA = test_pg_db_role_setting--1.0.sql
+
+PGFILEDESC = "test_pg_db_role_setting - tests for default GUC values stored in pg_db_role_settings"
+
+REGRESS = test_pg_db_role_setting
+
+# disable installcheck for now
+NO_INSTALLCHECK = 1
+# and also for now force NO_LOCALE and UTF8
+ENCODING = UTF8
+NO_LOCALE = 1
+
+ifdef USE_PGXS
+PG_CONFIG = pg_config
+PGXS := $(shell $(PG_CONFIG) --pgxs)
+include $(PGXS)
+else
+subdir = src/test/modules/test_pg_db_role_setting
+top_builddir = ../../../..
+include $(top_builddir)/src/Makefile.global
+include $(top_srcdir)/contrib/contrib-global.mk
+endif
diff --git a/src/test/modules/test_pg_db_role_setting/expected/test_pg_db_role_setting.out b/src/test/modules/test_pg_db_role_setting/expected/test_pg_db_role_setting.out
new file mode 100644
index 00000000000..b67b20d8357
--- /dev/null
+++ b/src/test/modules/test_pg_db_role_setting/expected/test_pg_db_role_setting.out
@@ -0,0 +1,120 @@
+CREATE EXTENSION test_pg_db_role_settings;
+ERROR:  extension "test_pg_db_role_settings" is not available
+DETAIL:  Could not open extension control file "/Users/smagen/projects/postgresql/env/master/src/tmp_install/Users/smagen/projects/postgresql/env/master/pgsql/share/extension/test_pg_db_role_settings.control": No such file or directory.
+HINT:  The extension must first be installed on the system where PostgreSQL is running.
+CREATE USER super_user SUPERUSER;
+CREATE USER regular_user;
+\c - regular_user
+-- successfully set a placeholder value
+SET test_pg_db_role_settings.superuser_param = 'aaa';
+-- module is loaded, the placeholder value is thrown away
+SELECT load_test_pg_db_role_settings();
+ERROR:  function load_test_pg_db_role_settings() does not exist
+LINE 1: SELECT load_test_pg_db_role_settings();
+               ^
+HINT:  No function matches the given name and argument types. You might need to add explicit type casts.
+SHOW test_pg_db_role_settings.superuser_param;
+ test_pg_db_role_settings.superuser_param 
+------------------------------------------
+ aaa
+(1 row)
+
+SHOW test_pg_db_role_settings.user_param;
+ERROR:  unrecognized configuration parameter "test_pg_db_role_settings.user_param"
+\c - regular_user
+-- fail, not privileges
+ALTER ROLE regular_user SET test_pg_db_role_settings.superuser_param = 'aaa';
+ERROR:  permission denied to set parameter "test_pg_db_role_settings.superuser_param"
+ALTER ROLE regular_user SET test_pg_db_role_settings.user_param = 'bbb';
+ERROR:  permission denied to set parameter "test_pg_db_role_settings.user_param"
+-- success for USER SET parameters
+ALTER ROLE regular_user SET test_pg_db_role_settings.superuser_param = 'aaa' USER SET;
+ALTER ROLE regular_user SET test_pg_db_role_settings.user_param = 'bbb' USER SET;
+SELECT * FROM pg_db_role_setting;
+ setdatabase | setrole |                                              setconfig                                               
+-------------+---------+------------------------------------------------------------------------------------------------------
+       16384 |       0 | {lc_messages=C,lc_monetary=C,lc_numeric=C,lc_time=C,bytea_output=hex,timezone_abbreviations=Default}
+           0 |   16386 | {test_pg_db_role_settings.superuser_param(u)=aaa,test_pg_db_role_settings.user_param(u)=bbb}
+(2 rows)
+
+\c - regular_user
+-- successfully set placeholders
+SHOW test_pg_db_role_settings.superuser_param;
+ test_pg_db_role_settings.superuser_param 
+------------------------------------------
+ aaa
+(1 row)
+
+SHOW test_pg_db_role_settings.user_param;
+ test_pg_db_role_settings.user_param 
+-------------------------------------
+ bbb
+(1 row)
+
+-- module is loaded, the placeholder value of superuser param is thrown away
+SELECT load_test_pg_db_role_settings();
+ERROR:  function load_test_pg_db_role_settings() does not exist
+LINE 1: SELECT load_test_pg_db_role_settings();
+               ^
+HINT:  No function matches the given name and argument types. You might need to add explicit type casts.
+SHOW test_pg_db_role_settings.superuser_param;
+ test_pg_db_role_settings.superuser_param 
+------------------------------------------
+ aaa
+(1 row)
+
+SHOW test_pg_db_role_settings.user_param;
+ test_pg_db_role_settings.user_param 
+-------------------------------------
+ bbb
+(1 row)
+
+\c - super_user
+SELECT load_test_pg_db_role_settings();
+ERROR:  function load_test_pg_db_role_settings() does not exist
+LINE 1: SELECT load_test_pg_db_role_settings();
+               ^
+HINT:  No function matches the given name and argument types. You might need to add explicit type casts.
+-- give the privilege to set SUSET param to the regular user
+GRANT SET ON PARAMETER test_pg_db_role_settings.superuser_param TO regular_user;
+\c - regular_user
+ALTER ROLE regular_user SET test_pg_db_role_settings.superuser_param = 'aaa';
+SELECT * FROM pg_db_role_setting;
+ setdatabase | setrole |                                              setconfig                                               
+-------------+---------+------------------------------------------------------------------------------------------------------
+       16384 |       0 | {lc_messages=C,lc_monetary=C,lc_numeric=C,lc_time=C,bytea_output=hex,timezone_abbreviations=Default}
+           0 |   16386 | {test_pg_db_role_settings.superuser_param=aaa,test_pg_db_role_settings.user_param(u)=bbb}
+(2 rows)
+
+\c - regular_user
+-- successfully set placeholders
+SHOW test_pg_db_role_settings.superuser_param;
+ test_pg_db_role_settings.superuser_param 
+------------------------------------------
+ aaa
+(1 row)
+
+SHOW test_pg_db_role_settings.user_param;
+ test_pg_db_role_settings.user_param 
+-------------------------------------
+ bbb
+(1 row)
+
+-- module is loaded, and placeholder values are succesfully set
+SELECT load_test_pg_db_role_settings();
+ERROR:  function load_test_pg_db_role_settings() does not exist
+LINE 1: SELECT load_test_pg_db_role_settings();
+               ^
+HINT:  No function matches the given name and argument types. You might need to add explicit type casts.
+SHOW test_pg_db_role_settings.superuser_param;
+ test_pg_db_role_settings.superuser_param 
+------------------------------------------
+ aaa
+(1 row)
+
+SHOW test_pg_db_role_settings.user_param;
+ test_pg_db_role_settings.user_param 
+-------------------------------------
+ bbb
+(1 row)
+
diff --git a/src/test/modules/test_pg_db_role_setting/meson.build b/src/test/modules/test_pg_db_role_setting/meson.build
new file mode 100644
index 00000000000..ffb50d830e0
--- /dev/null
+++ b/src/test/modules/test_pg_db_role_setting/meson.build
@@ -0,0 +1,35 @@
+# FIXME: prevent install during main install, but not during test :/
+
+test_oat_hooks_sources = files(
+  'test_pg_db_role_setting.c',
+)
+
+if host_system == 'windows'
+  test_oat_hooks_sources += rc_lib_gen.process(win32ver_rc, extra_args: [
+    '--NAME', 'test_pg_db_role_setting',
+    '--FILEDESC', 'test_pg_db_role_setting - tests for default GUC values stored in pg_db_role_settings',])
+endif
+
+test_pg_db_role_setting = shared_module('test_pg_db_role_setting',
+  test_pg_db_role_setting_sources,
+  kwargs: pg_mod_args,
+)
+testprep_targets += test_pg_db_role_setting
+
+install_data(
+  'test_pg_db_role_setting.control',
+  'test_pg_db_role_setting--1.0.sql',
+  kwargs: contrib_data_args,
+)
+
+tests += {
+  'name': 'test_pg_db_role_setting',
+  'sd': meson.current_source_dir(),
+  'bd': meson.current_build_dir(),
+  'regress': {
+    'sql': [
+      'test_pg_db_role_setting',
+    ],
+    'regress_args': ['--no-locale', '--encoding=UTF8'],
+  },
+}
diff --git a/src/test/modules/test_pg_db_role_setting/sql/test_pg_db_role_setting.sql b/src/test/modules/test_pg_db_role_setting/sql/test_pg_db_role_setting.sql
new file mode 100644
index 00000000000..4c2576cb9be
--- /dev/null
+++ b/src/test/modules/test_pg_db_role_setting/sql/test_pg_db_role_setting.sql
@@ -0,0 +1,55 @@
+CREATE EXTENSION test_pg_db_role_settings;
+CREATE USER super_user SUPERUSER;
+CREATE USER regular_user;
+
+\c - regular_user
+-- successfully set a placeholder value
+SET test_pg_db_role_settings.superuser_param = 'aaa';
+
+-- module is loaded, the placeholder value is thrown away
+SELECT load_test_pg_db_role_settings();
+
+SHOW test_pg_db_role_settings.superuser_param;
+SHOW test_pg_db_role_settings.user_param;
+
+\c - regular_user
+-- fail, not privileges
+ALTER ROLE regular_user SET test_pg_db_role_settings.superuser_param = 'aaa';
+ALTER ROLE regular_user SET test_pg_db_role_settings.user_param = 'bbb';
+-- success for USER SET parameters
+ALTER ROLE regular_user SET test_pg_db_role_settings.superuser_param = 'aaa' USER SET;
+ALTER ROLE regular_user SET test_pg_db_role_settings.user_param = 'bbb' USER SET;
+
+SELECT * FROM pg_db_role_setting;
+
+\c - regular_user
+-- successfully set placeholders
+SHOW test_pg_db_role_settings.superuser_param;
+SHOW test_pg_db_role_settings.user_param;
+
+-- module is loaded, the placeholder value of superuser param is thrown away
+SELECT load_test_pg_db_role_settings();
+
+SHOW test_pg_db_role_settings.superuser_param;
+SHOW test_pg_db_role_settings.user_param;
+
+\c - super_user
+SELECT load_test_pg_db_role_settings();
+-- give the privilege to set SUSET param to the regular user
+GRANT SET ON PARAMETER test_pg_db_role_settings.superuser_param TO regular_user;
+
+\c - regular_user
+ALTER ROLE regular_user SET test_pg_db_role_settings.superuser_param = 'aaa';
+
+SELECT * FROM pg_db_role_setting;
+
+\c - regular_user
+-- successfully set placeholders
+SHOW test_pg_db_role_settings.superuser_param;
+SHOW test_pg_db_role_settings.user_param;
+
+-- module is loaded, and placeholder values are succesfully set
+SELECT load_test_pg_db_role_settings();
+
+SHOW test_pg_db_role_settings.superuser_param;
+SHOW test_pg_db_role_settings.user_param;
diff --git a/src/test/modules/test_pg_db_role_setting/test_pg_db_role_setting--1.0.sql b/src/test/modules/test_pg_db_role_setting/test_pg_db_role_setting--1.0.sql
new file mode 100644
index 00000000000..1ed3d285c7e
--- /dev/null
+++ b/src/test/modules/test_pg_db_role_setting/test_pg_db_role_setting--1.0.sql
@@ -0,0 +1,7 @@
+/* src/test/modules/test_pg_db_role_setting/test_pg_db_role_setting--1.0.sql */
+
+-- complain if script is sourced in psql, rather than via CREATE EXTENSION
+\echo Use "CREATE EXTENSION test_pg_db_role_setting" to load this file. \quit
+
+CREATE FUNCTION load_test_pg_db_role_setting() RETURNS void
+  AS 'MODULE_PATHNAME' LANGUAGE C;
diff --git a/src/test/modules/test_pg_db_role_setting/test_pg_db_role_setting.c b/src/test/modules/test_pg_db_role_setting/test_pg_db_role_setting.c
new file mode 100644
index 00000000000..7d71cf305d7
--- /dev/null
+++ b/src/test/modules/test_pg_db_role_setting/test_pg_db_role_setting.c
@@ -0,0 +1,57 @@
+/*--------------------------------------------------------------------------
+ *
+ * test_pg_db_role_settings.c
+ *		Code for testing mandatory access control (MAC) using object access hooks.
+ *
+ * Copyright (c) 2015-2022, PostgreSQL Global Development Group
+ *
+ * IDENTIFICATION
+ *		src/test/modules/test_pg_db_role_settings/test_pg_db_role_settings.c
+ *
+ * -------------------------------------------------------------------------
+ */
+
+#include "postgres.h"
+
+#include "utils/guc.h"
+
+PG_MODULE_MAGIC;
+
+PG_FUNCTION_INFO_V1(load_test_pg_db_role_settings);
+
+static char *superuser_param;
+static char *user_param;
+
+/*
+ * Module load callback
+ */
+void
+_PG_init(void)
+{
+	DefineCustomStringVariable("test_pg_db_role_settings.superuser_param",
+							   "Sample superuser parameter.",
+							   NULL,
+							   &superuser_param,
+							   "superuser_param_value",
+							   PGC_SUSET,
+							   0,
+							   NULL, NULL, NULL);
+
+	DefineCustomStringVariable("test_pg_db_role_settings.user_param",
+							   "Sample user parameter.",
+							   NULL,
+							   &user_param,
+							   "user_param_value",
+							   PGC_USERSET,
+							   0,
+							   NULL, NULL, NULL);
+}
+
+/*
+ * Empty function, which is used just to trigger load of this module.
+ */
+Datum
+load_test_pg_db_role_settings(PG_FUNCTION_ARGS)
+{
+	PG_RETURN_VOID();
+}
diff --git a/src/test/modules/test_pg_db_role_setting/test_pg_db_role_setting.control b/src/test/modules/test_pg_db_role_setting/test_pg_db_role_setting.control
new file mode 100644
index 00000000000..9678cff376d
--- /dev/null
+++ b/src/test/modules/test_pg_db_role_setting/test_pg_db_role_setting.control
@@ -0,0 +1,7 @@
+# test_pg_db_role_setting extension
+comment = 'test_pg_db_role_setting - tests for default GUC values stored in pg_db_role_setting'
+default_version = '1.0'
+module_pathname = '$libdir/test_pg_db_role_setting'
+relocatable = true
+superuser = false
+trusted = true
-- 
2.24.3 (Apple Git-128)

